import itertools
import pathlib
from typing import (
    Callable,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
)
import spacy
from spacy.tokens import Doc, Span, SpanGroup
from spacy.util import ensure_path

from crtl.components.ann_vector_kb import ANNCandidateGenerator, MentionCandidate
from spacy.language import Language
from spacy.pipeline import TrainablePipe
from spacy.util import registry


Span.set_extension("candidate_mentions", default=[], force=True)


class CRTLTextLinker:
    """A class for linking text mentions to candidate entities."""

    NIL = "NIL"
    NA = "NA"

    def __init__(self, candidate_generator: ANNCandidateGenerator, k: int):
        """
        Initialize the CRTLTextLinker.

        Args:
            candidate_generator (ANNCandidateGenerator): The candidate generator for generating mention candidates.
        """
        self._k = k
        self._candidate_generator = candidate_generator

    def initialize(self):
        self._candidate_generator.initialize()

    def add_vocab_to_pipeline(self, nlp: spacy.Language):
        for concept in self._candidate_generator._concept_index:
            nlp.vocab.strings.add(str(concept.concept_id))

    def predict(
        self, mention_strings: Iterable[str]
    ) -> Iterable[Iterable[MentionCandidate]]:
        """
        Predict the candidates for the given mention strings.

        Args:
            mention_strings (Iterable[str]): Iterable of mention strings.

        Returns:
            Iterable[Iterable[MentionCandidate]]: Iterable of mention candidates for each mention string.
        """
        # mention_strings = list(mention_strings)
        candidates = self._candidate_generator(mention_strings, k=self._k)
        return candidates
        # return self._get_winning_links(batch_candidates)

    def to_disk(self, path: pathlib.Path):
        """
        Save the CRTLTextLinker to disk.

        Args:
            path (pathlib.Path): The path to save the model.
        """
        path = ensure_path(path)
        if not path.exists():
            path.mkdir()
        ann_candidate_generator_path = path / "candidate_generator"
        self._candidate_generator.to_disk(ann_candidate_generator_path)

    def from_disk(self, path: pathlib.Path):
        """
        Load the CRTLTextLinker from disk.

        Args:
            path (pathlib.Path): The path to load the model from.
            threshold: The k parameter for the CRTLTextLinker.
            ef_search: The ef_search parameter for the CRTLTextLinker.

        Returns:
            CRTLTextLinker: The loaded CRTLTextLinker object.
        """
        path = ensure_path(path)
        if not path.exists():
            path.mkdir()
        # This will receive the directory path + /my_component
        ann_candidate_generator_path = path / "candidate_generator"
        self._candidate_generator.from_disk(ann_candidate_generator_path)
        # how do we make this use an abstract candidate generator
        return self


class CRTLLinker(TrainablePipe):
    """A base class for CRTL linkers."""

    def __init__(
        self,
        crtl_text_linker: CRTLTextLinker,
        scorer: Optional[Callable] = None,
        cer_label_subset: Optional[List[str]] = None,
    ):
        """
        Initialize the CRTLLinker.

        Args:
            k (int): The k parameter for linking candidates.
            ef_search (int): The search parameter for the CRTLLinker.
            scorer (Optional[Callable]): Optional scorer function for scoring candidate entities.
        """
        self._text_linker = crtl_text_linker
        if cer_label_subset:
            self.cer_label_subset = set(cer_label_subset)
        else:
            self.cer_label_subset = None

        self.scorer = scorer

    def initialize(self, get_examples=None, nlp: spacy.Language = None):
        """
        Initialize the linker.

        Args:
            get_examples: spacy-required example getter.
            nlp: spacy-required language arg.
            candidate_generator_getter (Optional[Callable[[], ANNCandidateGenerator]]): Optional getter function for the candidate generator.
        """
        self._text_linker.initialize()
        self._text_linker.add_vocab_to_pipeline(nlp)

    def _get_spans(self, doc: Doc) -> Iterable[Span]:
        raise NotImplementedError

    def _set_spans(self, doc: Doc, new_spans: Iterable[Span]) -> None:
        raise NotImplementedError

    def predict(self, docs: Iterable[Doc]) -> Iterable[Iterable[MentionCandidate]]:
        """
        Predict the candidates for the given documents.

        Args:
            docs (Iterable[Doc]): Iterable of input documents.

        Returns:
            Iterable[Iterable[MentionCandidate]]: Iterable of mention candidates for each document.
        """
        if self.cer_label_subset:
            mention_strings = itertools.chain.from_iterable(
                (
                    (
                        span.text
                        for span in self._get_spans(doc)
                        if span.label_ in self.cer_label_subset
                    )
                    for doc in docs
                )
            )
        else:
            mention_strings = itertools.chain.from_iterable(
                ((span.text for span in self._get_spans(doc)) for doc in docs)
            )
        predictions = self._text_linker.predict(mention_strings)
        # print(list(predictions))
        # print(list(zip(mention_strings,list(prediction[0] for prediction in predictions))))
        return predictions

    def zip_with_offsets(
        self, iter_1: Iterable, iter_2: Iterable, offsets: Iterable[int]
    ) -> Iterable[Tuple]:
        """
        Zip one iterable with batches from another using a third offsets iterable.

        Args:
            iter_1 (Iterable): First iterable.
            iter_2 (Iterable): Second iterable.
            offsets (Iterable): Iterable of offsets.

        Returns:
            Iterable[Tuple]: Zipped iterable with offsets.
        """
        start = 0
        for iter_1_item, iter_1_offset in zip(iter_1, offsets):
            yield iter_1_item, itertools.islice(iter_2, start, start + iter_1_offset)
            start += iter_1_offset

    def _get_new_spans(
        self,
        docs: Iterable[Doc],
        concept_mention_predictions: Iterable[Iterable[MentionCandidate]],
    ) -> Iterable[Span]:
        """
        Get new spans with mention candidates added.

        Args:
            docs (Iterable[Doc]): Iterable of input documents.
            concept_mention_predictions (Iterable[Iterable[MentionCandidate]]): Iterable of mention candidates for each document.

        Returns:
            Iterable[Span]: Iterable of new spans with mention candidates added.
        """
        spans = itertools.chain.from_iterable(self._get_spans(doc) for doc in docs)
        new_spans = [
            Span(span.doc, span.start, span.end, label=span.label_) for span in spans
        ]
        if self.cer_label_subset:
            altered_spans = (
                span for span in new_spans if span.label_ in self.cer_label_subset
            )
        else:
            altered_spans = new_spans
        # print([new_span.text for new_span in altered_spans])
        for span, candidate_mentions in zip(altered_spans, concept_mention_predictions):
            span._.candidate_mentions = list(candidate_mentions)
        return new_spans

    def set_annotations(
        self,
        docs: Iterable[Doc],
        concept_mention_predictions: Iterable[Iterable[MentionCandidate]],
    ) -> None:
        """
        Set annotations in the documents with concept mention predictions.

        Args:
            docs (Iterable[Doc]): Iterable of input documents.
            concept_mention_predictions (Iterable[Iterable[MentionCandidate]]): Iterable of mention candidates for each document.
        """
        new_spans_batch = self._get_new_spans(docs, concept_mention_predictions)
        offsets = (len(self._get_spans(doc)) for doc in docs)
        for doc, new_spans in self.zip_with_offsets(docs, new_spans_batch, offsets):
            self._set_spans(doc, new_spans)

    def to_disk(self, path: pathlib.Path, exclude=tuple()):
        """
        Save the CRTLLinker to disk.

        Args:x
            path (pathlib.Path): The path to save the model.
            exclude (tuple): Unused argument.
        """
        self._text_linker.to_disk(path)

    def from_disk(self, path: pathlib.Path, exclude=tuple()):
        """
        Load the CRTLLinker from disk.

        Args:
            path (pathlib.Path): The path to the saved model.
            exclude (tuple): Unused argument.

        Returns:
            CRTLLinker: The loaded CRTLLinker.
        """
        self._text_linker = self._text_linker.from_disk(path)
        return self

    @property
    def is_trainable(self) -> bool:
        """
        Check if the component is trainable (base class property).

        Returns:
            bool: False
        """
        return False


class CRTLSpanLinker(CRTLLinker):
    def __init__(self, span_key, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._span_key = span_key

    def _get_spans(self, doc: Doc):
        """
        Get the spans from the document.

        Args:
            doc (Doc): The input document.

        Returns:
            Iterable[Span]: Iterable of spans.
        """
        return doc.spans[self._span_key]

    def _set_spans(self, doc: Doc, new_spans: Iterable[Span]):
        """
        Set the spans in the document.

        Args:
            doc (Doc): The input document.
            new_spans (Iterable[Span]): Iterable of new spans.
        """
        doc.spans[self._span_key] = SpanGroup(
            doc, name=self._span_key, spans=list(new_spans)
        )


class CRTLEntityLinker(CRTLLinker):
    def _get_spans(self, doc: Doc) -> Iterable[Span]:
        """
        Get the spans from the document.

        Args:
            doc (Doc): The input document.

        Returns:
            Iterable[Span]: Iterable of spans.
        """
        return doc.ents

    def _set_spans(self, doc: Doc, new_spans: Iterable[Span]):
        """
        Set the spans in the document.

        Args:
            doc (Doc): The input document.
            new_spans (Iterable[Span]): Iterable of new spans.
        """
        doc.set_ents(list(new_spans))


def get_setters_and_getters(
    ent_or_span: Literal["ent", "span"], span_key: Optional[str] = None
) -> Tuple[Callable[[Doc], Iterable[Span]], Callable[[Doc, Iterable[Span]], None]]:
    if ent_or_span == "ent":
        span_getter = lambda doc: doc.ents

        def span_setter(doc, spans):
            doc.set_ents(spans)

    else:
        assert span_key
        span_getter = lambda doc: doc.spans[span_key]

        def span_setter(doc, spans):
            doc.spans[span_key] = SpanGroup(doc, name=span_key, spans=list(spans))

    return span_getter, span_setter


@registry.misc("crtl_text_linker")
def create_crtl_text_linker(candidate_generator: ANNCandidateGenerator, k: int):
    return CRTLTextLinker(candidate_generator, k)


default_text_linker = (
    {
        "@misc": "crtl_text_linker",
        "candidate_generator": {
            "@misc": "ann_candidate_generator",
            "vectorizer": {
                "@misc": "transformer_vectorizer",
                "max_len": 50,
                "batch_size": 512,
                "device": "cpu",
            },
            "lkb_getter": {
                "@readers": "crtl.read_gzipped_jsonl_from_cached_path",
                "compressed_file_path": "s3://databrickstrenddata/crtl/lkb/trendnet_2.0.1.jsonl.gz",
            },
        },
        "k": 10,
    },
)


@Language.factory(
    name="trend_entity_linker",
    default_config={
        "text_linker": default_text_linker,
        "cer_label_subset": None,
        "scorer": {
            "@scorers": "crtl.score_accuracy_at_n",
            "ns": [1, 5, 10],
            "ent_or_span": "ent",
        },
    },
    default_score_weights={"rank_at_n_score": None},
)
def make_trend_entity_linker(
    nlp: Language,
    name: str,
    text_linker: CRTLTextLinker,
    scorer: Optional[Callable],
    cer_label_subset: Optional[List[str]],
):
    return CRTLEntityLinker(text_linker, scorer, cer_label_subset)


@Language.factory(
    name="trend_span_linker",
    default_config={
        "text_linker": default_text_linker,
        "span_key": "sc",
        "cer_label_subset": None,
        "scorer": {
            "@scorers": "crtl.score_accuracy_at_n",
            "ns": [1, 5, 10],
            "ent_or_span": "span",
            "span_key": "sc",
            "label_subset": None,
        },
    },
    default_score_weights={"rank_at_n_score": None},
)
def make_trend_span_linker(
    nlp: Language,
    name: str,
    text_linker: CRTLTextLinker,
    span_key: str,
    cer_label_subset: Optional[list[str]],
    scorer: Optional[Callable],
):
    return CRTLSpanLinker(span_key, text_linker, scorer, cer_label_subset)


if __name__ == "__main__":
    nlp = spacy.blank("en")
    doc_with_clinical_spans = nlp("The patient has a history of diabetes.")
    # nlp.add_pipe("trend_span_linker")
    # nlp.initialize()
    # nlp.from_disk("trend_span_linker")

    nlp = spacy.load("trend_span_linker")
    doc_with_clinical_spans.spans["sc"] = [
        doc_with_clinical_spans.char_span(4, 10, alignment_mode="expand")
    ]
    doc_with_clinical_spans = nlp(doc_with_clinical_spans)
    for span in doc_with_clinical_spans.spans["sc"]:
        print(span._.candidate_mentions[0].concept_id)
