import argparse
import datetime
import pathlib
from typing import Callable, Iterable, List, NamedTuple, Union
import numpy as np
import faiss
from crtl.lkb.lkb import LexicalKnowledgeBase

from crtl.components.vectorizer import (
    TransformerSpanVectorizer,
    Vectorizer,
)


class MentionCandidate(NamedTuple):
    """
    A data class representing a candidate entity that a mention may be linked to.
    Parameters
    ----------
    concept_id : str, required.
        The canonical concept id in the KB.
    similarities : List[float], required.
        The cosine similarities from the mention text to the alias in tf-idf space.
    """

    concept_id: str
    similarity: float


class FaissANNCandidateGenerator:
    """
    Candidate generator that uses Approximate Nearest Neighbor (ANN) search to find mention candidates.

    This class takes a Lexical Knowled
    ge Base (LKB), builds an ANN index of term vectors,
    and provides methods to generate mention candidates given input terms.

    Parameters:
        lkb (LexicalKnowledgeBase): The Lexical Knowledge Base containing term-to-concept mappings.
        term_index (list[str]): A list of terms corresponding to the vectors in the ANN index.
        ann_index: The ANN index used for nearest neighbor search.
        term_vectors: The term vectors used for similarity comparison.
        vectorizer (TfidfVectorizer): The TF-IDF vectorizer used to convert terms into vectors.

    """

    def __init__(
        self, vectorizer: Vectorizer, lkb_getter: Callable[[], LexicalKnowledgeBase]
    ):
        self._vectorizer = vectorizer
        self._lkb_getter = lkb_getter

    def initialize(self):
        """
        Initializes the ANNCandidateGenerator by creating vectors, building an approximate nearest neighbor index,
        and returning an instance of the ANNCandidateGenerator.
        """
        self._vectorizer.initialize()

        start_time = datetime.datetime.now()
        print(f"Creating  vectors...")
        vectors = self._get_normalized_vectors_from_concept_names()
        # np.save("umls_2024AA_concept_name_vectors.npy", vectors)
        vectors = np.load("umls_2024AA_concept_name_vectors.npy")
        print(f"Created vectors.")
        print(f"Creating index...")
        self._ann_index = self._create_index(vectors)
        print(f"Created index.")
        end_time = datetime.datetime.now()
        elapsed_time = end_time - start_time
        print(
            f"Creating candidate generator took {elapsed_time.total_seconds()} seconds"
        )
        return self

    def _get_cui_index(self, lkb):
        return sorted(lkb.cui_to_concept.keys())

    def _get_concept_names(self, lkb):
        cui_index = self._get_cui_index(lkb)
        return sorted([lkb.cui_to_concept[cui].canonical_name for cui in cui_index])

    def _get_normalized_vectors_from_concept_names(self):
        lkb = self._lkb_getter()
        concept_names = self._get_concept_names(lkb)
        return self._get_normalized_vectors(concept_names)

    def _get_unnormalized_vectors(self, terms):
        return self._vectorizer.transform(terms)

    def _get_normalized_vectors(self, terms):
        vectors = self._get_unnormalized_vectors(terms)
        faiss.normalize_L2(vectors)
        return vectors.astype(np.float32)

    def _create_index(self, vectors, m_parameter=100):
        d = vectors.shape[1]
        index = faiss.index_factory(
            d, "IVF65536_HNSW32,Flat", faiss.METRIC_L2
        )  # build the index
        # index = faiss.index_factory(d, "HNSW32", faiss.METRIC_L2)   # build the index
        index.hnsw.efConstruction = 200
        index.train(vectors)  # train on the dataset
        print(index.is_trained)
        index.add(vectors)  # add vectors to the index
        print(index.ntotal)
        return index

    def to_disk(self, path: Union[pathlib.Path, str]):
        path = pathlib.Path(path)
        path.mkdir(exist_ok=True)
        self._lkb.to_disk(path / "lkb")
        faiss.write_index(self._ann_index, path / "ann.index")
        self._vectorizer.to_disk(path / "vectorizer")

    def from_disk(self, path: pathlib.Path):
        self._ann_index = faiss.read_index(path / "ann.index")
        self._lkb = LexicalKnowledgeBase.from_disk(path / "lkb")
        self._cui_index = self._get_cui_index(self._lkb)
        self._vectorizer.from_disk(path / "vectorizer")
        return self

    def __call__(
        self, terms: Iterable[str], k: int
    ) -> Iterable[Iterable[MentionCandidate]]:
        terms = list(terms)
        vectors = self._get_normalized_vectors(terms)
        if len(terms) == 0:
            return []
        distances, neighbor_indices = self._ann_index.search(vectors, k)
        return self._get_mention_candidates(neighbor_indices, distances)

    def _get_mention_candidates(
        self, neighbor_indices: np.ndarray, distances: np.ndarray
    ) -> Iterable[List[MentionCandidate]]:
        for neighbors, distances in zip(neighbor_indices, distances):
            mention_candidates = []
            for neighbor_idx, distance in zip(neighbors, distances):
                cui = self._cui_index[neighbor_idx]
                concept_similarity = 1.0 - distance
                mention_candidates.append(
                    MentionCandidate(cui, similarity=concept_similarity)
                )
            yield mention_candidates


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "lkb_jsonl_path",
        type=pathlib.Path,
        help="Path to lkb .jsonl path",
    )
    parser.add_argument(
        "outpath",
        type=pathlib.Path,
        help="Path place the ",
    )

    args = parser.parse_args()
    vectorizer = TransformerSpanVectorizer(50, batch_size=512, device="mps").initialize(
        "cambridgeltl/SapBERT-from-PubMedBERT-fulltext",
        "cambridgeltl/SapBERT-from-PubMedBERT-fulltext",
    )  #
    candidate_generator = FaissANNCandidateGenerator(
        vectorizer, lambda: LexicalKnowledgeBase.from_disk(args.lkb_jsonl_path)
    ).initialize()
    candidate_generator.to_disk(args.outpath)
    terms = [
        "heart attack",
        "heart failure",
        "heart disease",
        "heart attacks",
        "hand",
        "HIE",
        "heart doctor",
        "liver doctor",
        "anxiety",
        "germ",
        "dryness",
    ]
    # candidate_generator = candidate_generator.initialize(lkb)
    for term, candidate_mentions in zip(terms, candidate_generator(terms, k=5)):

        for candidate in list(candidate_mentions)[:10]:

            print(
                term,
                "->",
                candidate_generator._lkb.cui_to_concept[
                    candidate.concept_id
                ].canonical_name,
                candidate.similarity,
            )
    # candidate_generator.to_disk(args.outpath)
