from abc import ABC, abstractmethod
from enum import Enum
from typing import Callable, TypeVar, Union
import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from transformers import BertModel
from transformers import AutoTokenizer
import pickle
import scipy
from crtl.lkb.lkb import LexicalKnowledgeBase
import torch
from tqdm import tqdm
from spacy.util import registry
import scipy.sparse
from transformers import AutoTokenizer

T = TypeVar("T", bound="Vectorizer")


class ArrayDataType(Enum):
    DENSE = "dense"
    SPARSE = "sparse"


class Vectorizer(ABC):
    datatype: ArrayDataType

    @abstractmethod
    def transform(self, texts) -> Union[np.ndarray, scipy.sparse.csr_matrix]:
        raise NotImplementedError

    @abstractmethod
    def initialize(self) -> T:
        raise NotImplementedError

    @abstractmethod
    def to_disk(self, path):
        raise NotImplementedError

    @abstractmethod
    def from_disk(self: T, path) -> T:
        raise NotImplementedError


class TransformerSpanVectorizer(Vectorizer):
    datatype = ArrayDataType.DENSE

    def __init__(
        self,
        max_len,
        batch_size=512,
        device="cuda",
        model_name="cambridgeltl/SapBERT-from-PubMedBERT-fulltext",
    ):
        self.max_len = max_len
        self.batch_size = batch_size
        self.device = device
        self._model_name = model_name

    def initialize(self):
        self.tokenizer = AutoTokenizer.from_pretrained(self._model_name)
        self.transformer_model = BertModel.from_pretrained(self._model_name)
        self.transformer_model.to(self.device)
        return self

    def to_disk(self, path):
        self.transformer_model.save_pretrained(path)
        self.tokenizer.save_pretrained(path)

    def from_disk(self, path):
        self.tokenizer = AutoTokenizer.from_pretrained(path)
        self.transformer_model = BertModel.from_pretrained(path)

    def _transform_batch(self, batch_of_texts) -> np.ndarray:
        # Tokenize texts in batch and move tensors to the same device as the model
        inputs = self.tokenizer(
            batch_of_texts,
            return_tensors="pt",
            padding=True,
            truncation=True,
            max_length=self.max_len,
        )
        inputs = {k: v.to(self.transformer_model.device) for k, v in inputs.items()}
        with torch.no_grad():  # Disable gradient calculation for inference
            outputs = self.transformer_model(**inputs)
        # Return the mean of the last hidden states and move it to CPU for further use
        return outputs.last_hidden_state.mean(dim=1).cpu().numpy()

    def transform(self, texts) -> np.ndarray:
        all_embeddings = []
        for i in tqdm(range(0, len(texts), self.batch_size)):
            batch_texts = texts[i : i + self.batch_size]
            embeddings = self._transform_batch(batch_texts)
            all_embeddings.extend(embeddings)
        return np.array(all_embeddings)


class TFIDFSpanVectorizer(Vectorizer):
    datatype = ArrayDataType.SPARSE

    def __init__(
        self,
        lkb_getter: Callable[[], LexicalKnowledgeBase],
        min_df=10,
        ngram_range=(3, 3),
    ) -> "TFIDFSpanVectorizer":
        self.lkb_getter = lkb_getter
        self._min_df = min_df
        self._ngram_range = ngram_range

    def initialize(self):
        self.vectorizer = TfidfVectorizer(
            analyzer="char_wb",
            ngram_range=self._ngram_range,
            min_df=self._min_df,
            dtype=np.float64,
        )
        self.vectorizer.fit_transform(self.lkb_getter().term_to_cuis.keys())
        return self

    def to_disk(self, path):
        pickle.dump(self.vectorizer, open(path, "wb"))

    def from_disk(self, path):
        self.vectorizer = pickle.load(open(path, "rb"))

    def transform(self, texts) -> scipy.sparse.csr_matrix:
        return self.vectorizer.transform(texts)


@registry.misc("tfidf_vectorizer")
def create_tfidf_vectorizer(lkb_getter, min_df=10, ngram_range=(3, 3)):
    return TFIDFSpanVectorizer(
        lkb_getter=lkb_getter, min_df=min_df, ngram_range=ngram_range
    )


@registry.misc("transformer_vectorizer")
def create_transformer_vectorizer(max_len, batch_size=512, device="cpu"):
    return TransformerSpanVectorizer(
        max_len=max_len, batch_size=batch_size, device=device
    )
