import argparse
from dataclasses import dataclass
from io import BytesIO
import shutil
from typing import Iterable, List, Optional
from spacy.tokens import DocBin
from spacy.lang.en import English
from spacy.tokens import Doc, Span, SpanGroup
import requests
import gzip

MEDMENTIONS_URL = "https://github.com/chanzuckerberg/MedMentions/raw/master/full/data/corpus_pubtator.txt.gz"


def download_medmentions(filename: str):
    response = requests.get(MEDMENTIONS_URL)

    # Make sure the request was successful
    response.raise_for_status()

    # Decompress the downloaded data and write it to a file
    with gzip.GzipFile(fileobj=BytesIO(response.content)) as f_in:
        with open(filename, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)


def parse_file_to_docs(filepath, outpath, tuis=None):
    docs = get_docs(filepath, tuis)
    DocBin(docs=docs).to_disk(outpath)
    print(f"Saved processed docs to {outpath}")
    print(f"First 5 documents and entities:")
    test_parsed_docs(outpath)
    return outpath


def get_docs(filepath, tuis: Optional[List[str]]):
    corpus = MedMentionCorpus.from_file(filepath)
    # Save the processed docs in a DocBin file
    # print([document.mentions for document in corpus.documents])
    if tuis:
        corpus = corpus.filter_tuis(tuis)
    return corpus.to_spacy_docs()


@dataclass
class Mention:
    pmid: str
    start_index: int
    end_index: int
    text_segment: str
    tuis: List[str]
    entity_id: str

    @classmethod
    def parse_line(cls, mention_line) -> "Mention":
        pmid, start_index, end_index, text_segment, tui, entity_id = mention_line.split(
            "\t"
        )
        start_index = int(start_index)
        end_index = int(end_index)
        tui = tui.split(",")
        return cls(pmid, start_index, end_index, text_segment, tui, entity_id)


@dataclass
class MedMentionDocument:
    title: str
    abstract: str
    mentions: List[Mention]

    @property
    def full_text(self):
        return self.title + " " + self.abstract

    @classmethod
    def parse_document(cls, document_str) -> Optional["MedMentionDocument"]:
        lines = document_str.split("\n")
        # Check if the current article has at least two lines to unpack
        if len(lines) < 2:
            print(
                f"Skipping an article due to insufficient lines. Content: {document_str}"
            )
            return None
        # Parse title and abstract
        title_line, abstract_line = lines[:2]
        title = "|".join(title_line.split("|")[2:])
        abstract = "|".join(abstract_line.split("|")[2:])
        # Parse entity lines
        mentions = [Mention.parse_line(line) for line in lines[2:]]
        return cls(title, abstract, mentions)

    def to_spacy_doc(self, tokenizer):
        spans = []
        doc = tokenizer(self.full_text)
        for mention in self.mentions:
            print(mention)
            start_token = None
            end_token = None
            for token in doc:
                if token.idx == mention.start_index:
                    start_token = token.i
                if token.idx + len(token) == mention.end_index:
                    end_token = token.i + 1  # +1 because span end indices are exclusive
                if start_token is not None and end_token is not None:
                    break
            if start_token is None or end_token is None:
                continue
            span = Span(
                doc, start_token, end_token, label="ENTITY", kb_id=mention.entity_id
            )
            spans.append(span)
        # print(len(spans))
        span_group = SpanGroup(doc, name=f"sc", spans=spans)
        doc.spans["sc"] = span_group
        # print(doc.spans["sc"])
        return doc


@dataclass
class MedMentionCorpus:
    documents: List[MedMentionDocument]

    @classmethod
    def from_file(cls, filepath) -> "MedMentionCorpus":
        with open(filepath, "r") as f:
            content = f.read()  # .replace(u"\u200e","")
            document_texts = content.split("\n\n")

        return cls(
            [
                document
                for document in [
                    MedMentionDocument.parse_document(document_text)
                    for document_text in document_texts
                ]
                if document
            ]
        )

    def filter_tuis(self, tuis: List[str]) -> "MedMentionCorpus":
        return MedMentionCorpus(
            [
                MedMentionDocument(
                    document.title,
                    document.abstract,
                    [
                        mention
                        for mention in document.mentions
                        if any(tui in tuis for tui in mention.tuis)
                    ],
                )
                for document in self.documents
            ]
        )

    def to_spacy_docs(self) -> Iterable[Doc]:
        nlp = English()
        tokenizer = nlp.tokenizer
        return [document.to_spacy_doc(tokenizer) for document in self.documents]


def test_parsed_docs(doc_bin_path):
    nlp = English()
    # Create a new DocBin instance
    doc_bin = DocBin()

    # Load the DocBin file from disk
    doc_bin = doc_bin.from_disk(doc_bin_path)

    # Load the stored documents into a list
    docs = list(doc_bin.get_docs(nlp.vocab))

    for doc in docs[:5]:
        print(doc.spans["sc"])


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("output_path", help="Path to place the resulting spacy DocBin.")
    parser.add_argument(
        "--medmentions_path",
        help="Path to the medmentions file. If not specified, will download from the MedMentions github repo.",
    )
    parser.add_argument(
        "--tuis", help="Space-Separated list of tuis to include", nargs="+"
    )
    args = parser.parse_args()

    if not args.medmentions_path:
        medmentions_path = "medmentions.txt"
        download_medmentions(medmentions_path)
    else:
        medmentions_path = args.medmentions_path

    parse_file_to_docs(medmentions_path, args.output_path, args.tuis)


if __name__ == "__main__":
    main()
